//
// p r i n t f . c
//
// Original printf module for stand-alone systems
//
// Targets: x86-GCC and ARM-GCC (WITH frame)
// NOTE: Requires external "void putchar(int)" function.
//
// May 21, 2003 by Wataru Nishida (http://www.skyfree.org)
//

#ifdef DEBUG
#include <stdio.h>
#endif

static void sputchar(int);
static void doprint(void (*)(int), const char*, long*);
static void printn(void (*)(int), unsigned long, int, int, int);

// Local variable

static char* strptr;	// String buffer pointer

// Functions

#ifdef DEBUG
int xsprintf(char* buf, const char* fmt, ...) {
#else
int sprintf(char* buf, const char* fmt, ...) {
#endif
  long* argptr = (long *) (&fmt + 1);	// argptr points third argument
  
  strptr = buf;				// Reset global buffer pointer
  doprint(sputchar, fmt, argptr);
  *strptr = '\0';			// Null termination
  return 0;
 }

#ifndef DEBUG
void putchar(int);			// We need external putchar()

int printf(const char* fmt, ...) {
  long* argptr = (long *) (&fmt + 1);	// argptr points third argument
  
  doprint(putchar, fmt, argptr);
  return 0;
 }
#endif

static
void sputchar(int ch) {
  *strptr++ = (char) ch;		// Save character to buffer
 }

static
void doprint(void (*put)(int), const char* fmt, long* argptr) {
  char* p;
  char  ch;
  unsigned long val;
  int digits;
  int zsup;

  for (;;) {
    while ((ch = *fmt++) != '%') {	// Search conversion specifier
      if (ch == '\0') return;
      (*put)((int) ch);
     }

    zsup = 1;				// Default is zero suppless mode
    if (*fmt == '0') {
      zsup = 0; fmt++;			// Zero(es) will be supplied
     }

    digits = 0;				// Analyze width specifier
    while (*fmt >= '0' && *fmt <= '9') {
      digits *= 10;
      digits += *fmt++ - '0';
     }

    switch (ch = *fmt++) {		// Analyze specifier
      case 'c':				// " Character "
        ch = (char) *argptr++;
        (*put)((int) ch);
        break;

      case 's':				// " String "
        p = (char*) *argptr++;
        while ((ch = *p++))
          (*put)((int) ch);
        break;

      case 'b':				// " Binary "
        val = (unsigned long) *argptr++;
        printn(put, val, 2, digits, zsup);
        break;

      case 'd':				// " Decimal "
        val = (unsigned long) *argptr++;
        printn(put, val, 10, digits, zsup);
        break;

      case 'o':				// " Octal "
        val = (unsigned long) *argptr++;
        printn(put, val, 8, digits, zsup);
        break;

      case 'x':				// " Hexadecimal "
        val = (unsigned long) *argptr++;
        printn(put, val, 16, digits, zsup);
        break;

      default:
        (*put)((int) '%'); (*put)((int) ch);
        break;
     }
   }
 }

static void
printn(void (*put)(int), unsigned long val, int base, int digits, int zsup) {
  char  buf[ 33 ];
  char* p;
  int cnt;

  if (digits > 32)
    digits = 32;
  else if (digits <= 0)
    digits = -1;

  p = buf;
  cnt = 0;
  do {
    if (cnt == 33 || cnt == digits) return;
    *p++ = "0123456789ABCDEF"[val % base];
    cnt++;
   } while (val /= base);

  while (cnt < digits) {
    if (zsup) 
      (*put)((int) ' ');
    else
      (*put)((int) '0');
    cnt++;
   }

  do {
    (*put)((int) (*--p));
   } while (p > buf);
 }

#ifdef DEBUG
int main() {
  char buf[ 128 ];

  xsprintf(buf, "%s mode%c %b/%o/%d/%x", "Zero-suppless", ':',
           251,252,253,254);
  printf("%s\n", buf);
  xsprintf(buf, "%s mode%c %010b/%06o/%06d/%06x", "Non zero-suppless", ':',
           251,252,253,254);
  printf("%s\n", buf);
  return 0;
 }
#endif
