//
// m a k e w a v e . c
//
// Create a WAVE file
//
// June 22, 2003 by Wataru Nishida (http://www.skyfree.org)
//

#include <unistd.h>	// write(), close()
#include <fcntl.h>	// open(), O_RDWR
#include <stdlib.h>	// atof()
#include <math.h>	// sin()
#include <stdio.h>	// printf()

// Constant definitions

#define WAVFRQ	44100	// WAVE frequency
#define SNDLEN	3	// WAVE length (seconds)
#define MAXAMP	20000	// Maximum WAVE amplitude
#define PI	3.14159	// The circular constant

// Define unions which manipulate little-endian

typedef union {
  int value;
  unsigned char image[ sizeof(int) ];
 } lint;

typedef union {
  short int value;
  unsigned char image[ sizeof(short int) ];
 } lshort;

// RIFF chunk definition

typedef struct {
  char id[ 4 ];		// "RIFF" group ID
  lint size;		// Chunk size (total - 8)
  char type[ 4 ];	// Form type "WAVE"
 } riff_chunk __attribute__((packed));

// Format subchunk definition

typedef struct {
  char   id[ 4 ];	// "fmt" ID
  lint   size;		// fmt chunk size
  lshort format;	// Audio format type (PCM = 1)
  lshort channels;	// Number of channels: Mono = 1, stereo = 2
  lint   sample_rate;	// Sample rate
  lint   xfer_rate;	// Transfer rate (sample_rate * channels * bits / 8)
  lshort bytes;		// Bytes / sample (bits * channels / 8)
  lshort bits;		// Bits / sample
 } fmt_chunk __attribute__((packed));

// Data subchunk definition

typedef struct {
  char id[ 4 ];		// "data" ID
  lint size;		// data chunk size
 } data_chunk __attribute__((packed));

// Global variables

riff_chunk riff_hdr;			// RIFF chunk header
fmt_chunk  fmt_hdr;			// Format subchunk header
data_chunk data_hdr;			// Data subchunk header
lshort     wave_data[ WAVFRQ*SNDLEN ];	// Wave data holder

// Save 4 byte integer with little-endian format

void save_little_int(lint* target, int val) {
  target->image[ 0 ] = (val & 0x000000FF);
  target->image[ 1 ] = (val & 0x0000FF00) >> 8;
  target->image[ 2 ] = (val & 0x00FF0000) >> 16;
  target->image[ 3 ] = (val & 0xFF000000) >> 24;
 }

// Save 2 byte short integer with little-endian format

void save_little_short(lshort* target, int val) {
  target->image[ 0 ] = (val & 0x000000FF);
  target->image[ 1 ] = (val & 0x0000FF00) >> 8;
 }

// Setup RIFF header

void setup_riff(riff_chunk* r, int size) {
  r->id[ 0 ] = 'R';
  r->id[ 1 ] = 'I';
  r->id[ 2 ] = 'F';
  r->id[ 3 ] = 'F';

  save_little_int(&r->size, size);

  r->type[ 0 ] = 'W';
  r->type[ 1 ] = 'A';
  r->type[ 2 ] = 'V';
  r->type[ 3 ] = 'E';
 }

// Setup fmt header

void setup_fmt(fmt_chunk* f) {
  f->id[ 0 ] = 'f';
  f->id[ 1 ] = 'm';
  f->id[ 2 ] = 't';
  f->id[ 3 ] = ' ';

  save_little_int(&f->size, 16);	// fmt chunk is 16 bytes
  save_little_short(&f->format, 1);	// Linear quantization (PCM)
  save_little_short(&f->channels, 1);	// Monoral 1ch
  save_little_int(&f->sample_rate, WAVFRQ);	// 44100 Hz
  save_little_int(&f->xfer_rate, WAVFRQ*1);	// 44100 bytes * 1 ch
  save_little_short(&f->bytes, 2);	// 16 bits * 1 ch = 2 bytes
  save_little_short(&f->bits, 16);	// 16 bits / sample
 }

// Setup data header

void setup_data(data_chunk* d, int size) {
  d->id[ 0 ] = 'd';
  d->id[ 1 ] = 'a';
  d->id[ 2 ] = 't';
  d->id[ 3 ] = 'a';

  save_little_int(&d->size, size);
 }

int main(int argc, char* argv[]) {
  double freq1 = 0, freq2 = 0, freq3 = 0, freq4 = 0, freq5 = 0, amp;
  int fd, ret, i;
  double data;

  if (argc < 3 || argc > 7) {
    printf("makewave filename freq1 [ freq2 [ freq3 [ freq4 [ freq5 ]]]]\n");
    return 1;
   }

  // Read frequencies

  freq1 = atof(argv[ 2 ]);
  if (argc >= 4)
    freq2 = atof(argv[ 3 ]);
  if (argc >= 5)
    freq3 = atof(argv[ 4 ]);
  if (argc >= 6)
    freq4 = atof(argv[ 5 ]);
  if (argc == 7)
    freq5 = atof(argv[ 6 ]);

  // Adjust wave amplitude

  argc -= 2;
  switch (argc) {
    case 1:
      amp = MAXAMP;
      break;

    case 2:
      amp = MAXAMP * 0.55;
      break;

    case 3:
      amp = MAXAMP * 0.4;
      break;

    case 4:
      amp = MAXAMP * 0.3;
      break;

    case 5:
      amp = MAXAMP * 0.25;
      break;
   }

  fd = open(argv[ 1 ], O_CREAT | O_RDWR);
  if (fd < 0) return 1;

  setup_riff(&riff_hdr, 36 + WAVFRQ * SNDLEN);
  ret = write(fd, &riff_hdr, sizeof(riff_chunk));
  if (ret != sizeof(riff_chunk)) return 1;

  setup_fmt(&fmt_hdr);
  ret = write(fd, &fmt_hdr, sizeof(fmt_chunk));
  if (ret != sizeof(fmt_chunk)) return 1;

  setup_data(&data_hdr, WAVFRQ * SNDLEN);
  ret = write(fd, &data_hdr, sizeof(data_chunk));
  if (ret != sizeof(data_chunk)) return 1;

  for (i = 0; i < WAVFRQ * SNDLEN; i++) {
    data = amp * sin(2 * PI * freq1 / WAVFRQ * i);
    if (freq2)
      data += amp * sin(2 * PI * freq2 / WAVFRQ * i);
    if (freq3)
      data += amp * sin(2 * PI * freq3 / WAVFRQ * i);
    if (freq4)
      data += amp * sin(2 * PI * freq4 / WAVFRQ * i);
    if (freq5)
      data += amp * sin(2 * PI * freq5 / WAVFRQ * i);
    save_little_short(&wave_data[ i ], (short) data);
   }

  ret = write(fd, &wave_data, WAVFRQ * SNDLEN);
  if (ret != (WAVFRQ * SNDLEN)) return 1;

  close(fd);
  printf("WAVE file \"%s\"\n", argv[ 1 ]);
  printf("Sampling rate = %d Hz : Sound length = %d sec\n", WAVFRQ, SNDLEN);
  printf("Freq1 = %f\nFreq2 = %f\nFreq3 = %f\n", freq1, freq2, freq3);
  printf("Freq4 = %f\nFreq5 = %f\n", freq4, freq5);
  return 0;
 }
