//
// w a v e 2 g b a . c
//
// Convert a WAVE file to GBA compatible format
//
// NOTICE: WAVE file must be PCM, monoral, 8bits format!
//
// July 26, 2003 by Wataru Nishida (http://www.skyfree.org)
//

#include <unistd.h>	// read(), close(), lseek(), SEEK_CUR
#include <fcntl.h>	// open(), O_RDONLY
#include <stdio.h>	// fprintf(), stdout, stderr
#include <stdlib.h>	// malloc(), exit()
#include <string.h>	// strncmp()

// Define unions which manipulate little-endian

typedef union {
  int value;
  unsigned char image[ sizeof(int) ];
 } lint;

typedef union {
  short int value;
  unsigned char image[ sizeof(short int) ];
 } lshort;

// RIFF chunk definition

typedef struct {
  char  id[ 4 ];	// "RIFF" group ID
  lint  size;		// Chunk size (total - 8)
  char  type[ 4 ];	// Form type "WAVE"
 } riff_chunk __attribute__((packed));

// Format subchunk definition

typedef struct {
  char   id[ 4 ];	// "fmt " ID
  lint   size;		// fmt chunk size
  lshort format;	// Audio format type (PCM = 1)
  lshort channels;	// Number of channels: Mono = 1, stereo = 2
  lint   sample_rate;	// Sample rate
  lint   xfer_rate;	// Transfer rate (sample_rate * channels * bits / 8)
  lshort bytes;		// Bytes / frame (bits * channels / 8)
  lshort bits;		// Bits / sample
 } fmt_chunk __attribute__((packed));

// Subchunk header definition

typedef struct {
  char id[ 4 ];		// ID
  lint size;		// Subchunk size
 } sub_chunk __attribute((packed));

// Global variables

riff_chunk riff_hdr;			// RIFF chunk header
fmt_chunk  fmt_hdr;			// Format subchunk header
sub_chunk  sub_hdr;			// Subchunk header

// Error handler

void die(const char* msg) {
  fprintf(stderr, "%s\n", msg);
  exit(1);
 }

// Load 4 byte little-endian integer

int load_little_int(lint* target) {
  int val;

  val  = target->image[ 0 ];
  val += target->image[ 1 ] << 8;
  val += target->image[ 2 ] << 16;
  val += target->image[ 3 ] << 24;
  return val;
 }

// Load 2 byte little-endian short integer

int load_little_short(lshort* target) {
  int val;

  val  = target->image[ 0 ];
  val += target->image[ 1 ] << 8;
  return val;
 }

// Read RIFF header

int read_riff(riff_chunk* r) {
  if (strncmp(r->id, "RIFF", 4))
    die("Illegal RIFF header");
  if (strncmp(r->type, "WAVE", 4))
    die("Non \"WAVE\" type");
  return(load_little_int(&r->size));
 }

// Read fmt header

void read_fmt(fmt_chunk* f, int* size, int* sample, int* bits) {

  if (strncmp(f->id, "fmt ", 4))
    die("Non \"fmt\" subchank");
  *size = load_little_int(&f->size);
  if (*size < 16)
    die("Illegal fmt subchank size");
  if (load_little_short(&f->format) != 1)
    die("Audio type must be PCM");
  if (load_little_short(&f->channels) != 1)
    die("Currently, wave2gba supports only monoral source");
  *sample = load_little_int(&f->sample_rate);	// Read sample rate
  *bits   = load_little_short(&f->bits);	// Read sample length
  if (*bits != 8)
    die("Audio format must be 8bits");
 }

int main(int argc, char* argv[]) {
  int fd, ret, size, bytes, sample, bits, i, data;
  void* buf;
  unsigned char* ptr;

  // Check argument

  if (argc != 2)
    die("Usage: wave2gba wave_file");

  // Open source file

  fd = open(argv[ 1 ], O_RDONLY);
  if (fd < 0) die("File open failure");

  // Check RIFF header

  ret = read(fd, &riff_hdr, sizeof(riff_chunk));
  if (ret != sizeof(riff_chunk)) die("\"RIFF\" heder read error");
  ret = read_riff(&riff_hdr);

  // Read format header

  ret = read(fd, &fmt_hdr, sizeof(fmt_chunk));
  if (ret != sizeof(fmt_chunk)) die("\"fmt\" header read error");
  read_fmt(&fmt_hdr, &size, &sample, &bits);
  if (size > (sizeof(fmt_chunk) - 8))
    ret = lseek(fd, size - sizeof(fmt_chunk) + 8, SEEK_CUR);

  // Search "data" subchunk header

  while (1) {
    ret = read(fd, &sub_hdr, sizeof(sub_chunk));
    if (ret != sizeof(sub_chunk)) die("Subchunk header read error");
    bytes = load_little_int(&sub_hdr.size);
    if (strncmp(sub_hdr.id, "data", 4) == 0) break;
    lseek(fd, bytes, SEEK_CUR);
   }

  // Read sound data

  buf = malloc(bytes);
  if (buf == NULL) die("Insufficient memory");
  ptr = (char*) buf;
  ret = read(fd, buf, bytes);
  if (ret != bytes) die("Sound data read error");
  close(fd);

  // Output C source file

  printf("// GBA sound file created by wave2gba\n\n");
  printf("const char wave_comment[] = \"%s\";\n", argv[ 1 ]);
  printf("int  wave_freq = %d;\n", sample);
  printf("int  wave_size = %d;\n", bytes);
  printf("char wave_data[] = {\n  ");
  for (i = 0; i < bytes; i++) {
    data = (*ptr++);
    // data *= 3; data >>= 2;
    printf("%d, ", data - 128);
    if ((i % 10) == 9) printf("\n  ");
   }
  printf(" };\n");

  return 0;
 }
